
import requests
# import typer
from rich import print


import os 
import json


from . import get_token, get_project_id, get_org_id
from pureml.utils.constants import BASE_URL, PATH_MODEL_DIR
from pureml import save_model, load_model
from urllib.parse import urljoin
import joblib
from pureml.utils.hash import check_hash_status_model

# app = typer.Typer()


# @app.command()
def list():
    '''This function will return a list of all the models in the project
    
    Returns
    -------
        A list of all the models in the project
    
    '''

    user_token = get_token()
    org_id = get_org_id()
    project_id = get_project_id()

    url_path_1 = '{}/project/{}/models'.format(org_id, project_id)
    url = urljoin(BASE_URL, url_path_1)
    
    data = {"project_id": project_id}

    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': 'Bearer {}'.format(user_token)
    }


    response = requests.get(url,data=data, headers=headers)
    
    if response.status_code == 200:
        # print(f"[bold green]Obtained list of models")
        
        response_text = response.json()
        model_list = response_text['data']
        # print(model_list)

        return model_list
    else:
        print(f"[bold red]Unable to obtain the list of models!")
        
    return




def check_model_hash(hash: str, name:str):

    user_token = get_token()
    org_id = get_org_id()
    project_id = get_project_id()


    url_path_1 = '{}/project/{}/model/{}/hash_status'.format(org_id, project_id, name)
    url = urljoin(BASE_URL, url_path_1)


    headers = {
        'Authorization': 'Bearer {}'.format(user_token)
    }


    data = {'hash': hash, 'project_id':project_id}
    response = requests.post(url, data=data, headers=headers)

    hash_exists = False

    if response.status_code == 200:
        # print(f"[bold green]Model hash verified!")

        hash_exists = response.json()['data']

        # if response.text == True:       #Change this logic accordingly
        #     hash_exists = True

    # else:
    #     print(f"[bold red]Model has not been verified!")
        # print(response.status_code)
        # print(response.text)

    return hash_exists




# @app.command()
def register(model, name:str) -> str:
    ''' The function takes in a model, a name and a version and saves the model locally, then uploads the
    model to the PureML server
    
    Parameters
    ----------
    model
        The model you want to register
    name : str
        The name of the model.
    
    '''
        
    user_token = get_token()
    org_id = get_org_id()
    project_id = get_project_id()


    model_file_name = '.'.join([name, 'pkl'])
    model_path = os.path.join(PATH_MODEL_DIR, model_file_name)

    os.makedirs(PATH_MODEL_DIR, exist_ok=True)
    
    save_model(model, name, model_path=model_path)

    # name_with_ext = '.'.join([name, 'pkl'])
    # model_path = os.path.join(PATH_MODEL_DIR, name_with_ext)

    model_exists_locally, model_hash =  check_hash_status_model(file_path = model_path, name=name, item_key='model')
    model_exists_remote = check_model_hash(hash=model_hash, name=name)


    if model_exists_remote:
        print(f"[bold red]Model already exists. Not registering a new version!")
        return True, model_hash, 'latest'
    else:        
        url_path_1 = '{}/project/{}/model/register'.format(org_id, project_id)
        url = urljoin(BASE_URL, url_path_1)


        headers = {
            'Authorization': 'Bearer {}'.format(user_token)
        }


        files = {'file': (model_file_name, open(model_path, 'rb'))}

        data = {'name': name, 'project_id':project_id, 'hash':model_hash}
        response = requests.post(url, files=files, data=data, headers=headers)

        if response.status_code == 200:
            print(f"[bold green]Model has been registered!")

            model_version = response.json()['data'][0]['version']

            return True, model_hash, model_version

        else:
            print(f"[bold red]Model has not been registered!")
    
    
        return False, model_hash, None




def details(name:str, version:str='latest'):
    '''It fetches the details of a model.
    
    Parameters
    ----------
    name : str
        The name of the model
    version: str
        The version of the model
    Returns
    -------
        The details of the model.
    
    '''
    
    user_token = get_token()
    org_id = get_org_id()
    project_id = get_project_id()

    url_path_1 = '{}/project/{}/model/{}/{}/details'.format(org_id, project_id, name, version)
    url = urljoin(BASE_URL, url_path_1)

    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': 'Bearer {}'.format(user_token)
    }
    

    response = requests.get(url, headers=headers)
    # print(response.url)
    # print(response.text)

    if response.status_code == 200:
        # print(f"[bold green]Model details have been fetched")
        response_text = response.json()
        model_details = response_text['data'][0]
        # print(model_details)

        return model_details

    else:
        print(f"[bold red]Model details have not been found")
        return 



# @app.command()
def fetch(name:str, version:str='latest'):
    '''This function fetches a model from the server and returns it as a `Model` object
    
    Parameters
    ----------
    name : str, optional
        The name of the model you want to fetch.
    version: str
        The version of the model
    
    Returns
    -------
        The model is being returned.
    
    '''

    user_token = get_token()
    org_id = get_org_id()
    project_id = get_project_id()



    model_details = details(name=name, version=version)

    if model_details is None:
        print(f"[bold red]Unable to fetch Model")
        return


    # model_location = model_details['location']
    # model_url = 'https://{}'.format(model_location)
    model_url = model_details['location']
    
    model_url = model_url.replace('https://pureml-registry.67f23f4479798ac96c01212517a90146.r2.cloudflarestorage.com', 
                                        'https://pub-072ac07f18cd4246bd5c879e7a9df94e.r2.dev')
    # print(model_url)

    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': 'Bearer {}'.format(user_token)
    }
    
    # print('url', model_url)

    # response = requests.get(model_url, headers=headers)

    response = requests.get(model_url)

    if response.status_code == 200:
        model_bytes = response.content
        open('temp_model.pure', 'wb').write(model_bytes)

        model = load_model(model_path='temp_model.pure')


        # print(f"[bold green]Model has been fetched")
        return model
    else:
        print(f"[bold red]Unable to fetch Model")
        # print(response.status_code)
        # print(response.text)
        # print(response.url)
        return 



# @app.command()
def delete(name:str, version:str='latest') -> str:
    ''' This function deletes a model from the project
    
    Parameters
    ----------
    name : str
        The name of the model you want to delete
    version : str
        The version of the model to delete.
    
    '''

    user_token = get_token()
    org_id = get_org_id()
    project_id = get_project_id()


    url_path_1 = '{}/project/{}/model/{}/delete'.format(org_id, project_id, name)
    url = urljoin(BASE_URL, url_path_1)

    
    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': 'Bearer {}'.format(user_token)
    }

    data = {'version':version}
    
    
    response = requests.delete(url, headers=headers, data=data)


    if response.status_code == 200:
        print(f"[bold green]Model has been deleted")
        
    else:
        print(f"[bold red]Unable to delete Model")

    return response.text


# @app.command()
def serve_model():
    pass



# if __name__ == "__main__":
#     app()



from .config import load_config, save_config
from .constants import PATH_CONFIG, PATH_PREDICT, PATH_PREDICT_DIR, PATH_PREDICT_REQUIREMENTS
from .hash import generate_hash_for_dict, generate_hash_for_function
from .source_code import get_source_code
import os
import shutil

def add_load_data_to_config(name, func=None, hash=''):

    config = load_config()
    code = ''
    if func is not None:
        try:
            code = get_source_code(func)
            hash = generate_hash_for_function(func)
        except Exception as e:
            print('Unable to get load_data source code')
            print(e)

    config['load_data'] = {
                            'name' : name,
                            'hash' : hash,
                            'code' : code
                            }

    save_config(config=config)




def add_transformer_to_config(name, func=None, hash='', parent=None):


    config = load_config()
    # print(config)
    position = len(config['transformer']) + 1

    if parent is None:
        if position == 1 :

            if len(config['load_data']) !=0:
                parent = config['load_data']['name']

    
        else:
            transformer_previous = config['transformer'][position-1]
            parent = transformer_previous['name']



    code = ''
    if func is not None:
        try:
            code = get_source_code(func)
            hash = generate_hash_for_function(func)
        except Exception as e:
            print('Unable to get transformer source code')
            print(e)

    config['transformer'][position] = {
                                        'name' : name,
                                        'hash' : hash,
                                        'parent': parent,
                                        'code': code                                      
                                        }
    # print('saveing configuration for ', name)
    save_config(config=config)


def add_dataset_to_config(name, func=None, hash='', version='', parent=None):

    config = load_config()


    if parent is None:

        if len(config['transformer']) !=0:
            config_transformer = config['transformer']
            transformer_last = list(config_transformer.values())[-1]
            parent = transformer_last['name']



    code = ''
    if func is not None:
        try:
            code = get_source_code(func)
            hash = generate_hash_for_function(func)
        except Exception as e:
            print('Unable to get dataset source code')
            print(e)

    config['dataset'] = {
                        'name' : name,
                        'hash' : hash,
                        'version': version,
                        'parent' : parent ,
                        'code': code                                            
                        }



    save_config(config=config)
    


def add_model_to_config(name, func=None, hash='', version=''):
    # name = ''
    # hash = ''
    # version = ''

    config = load_config()

    code = ''
    if func is not None:
        try:
            code = get_source_code(func)
            hash = generate_hash_for_function(func)
        except Exception as e:
            print('Unable to get model source code')
            print(e)

    #Empty hash is passed to create the empty model with just model name the first time
    #Complete hash is passed to create the model with all the details in the second time
    if hash == '':
        position = len(config['model']) + 1
        
        config['model'][position] = {
                                        'name' : name,
                                        'hash' : hash,
                                        'version': version,
                                        'code': code
                                        }
    else:
        position = len(config['model'])
        model_name_position = config['model'][position]['name']
        if model_name_position == name:        
            config['model'][position]['hash'] = hash
            config['model'][position]['version'] = version
            config['model'][position]['code'] = code


    save_config(config=config)


def add_metrics_to_config(values, model_name=None, model_version=None, func=None):
    config = load_config()

    if model_name is None:
        model_name, model_version, model_hash = get_model_latest(config=config)



    if len(config['metrics']) != 0:
        metric_values = config['metrics']['values']
        metric_values.update(values)
    else:
        metric_values = values

    hash = generate_hash_for_dict(values=metric_values)


    config['metrics'].update({
                            'values' : metric_values,
                            'hash' : hash,
                            'model_name' : model_name,
                            'model_version' : model_version
                        })

    save_config(config=config)


def load_metrics_from_config():

    config = load_config()
    try:
        metrics = config['metrics']['values']
    except Exception as e:
        # print(e)
        print('No metrics are found in config')
        metrics = {}

    return metrics



def add_params_to_config(values, model_name=None, model_version=None, func=None):
    config = load_config()
    
    if model_name is None:
        model_name, model_version, model_hash = get_model_latest(config=config)



    if len(config['params']) != 0:
        param_values = config['params']['values']
        param_values.update(values)
    else:
        param_values = values

    hash = generate_hash_for_dict(values=param_values)


    config['params'].update({
                            'values' : param_values,
                            'hash' : hash,
                            'model_name' : model_name,
                            'model_version' : model_version
                        })

    save_config(config=config)




def load_params_from_config():

    config = load_config()
    try:
        metrics = config['params']['values']
    except Exception as e:
        # print(e)
        print('No params are found in config')
        metrics = {}

    return metrics




def add_artifacts_to_config(name, values, func):
    hash = ''
    version = ''
    config = load_config()
    
    model_name, model_version = get_model_latest(config=config)

    position = len(config['artifacts']) + 1
    config['artifacts'][position] = {
                                        'name' : name,
                                        'hash' : hash,
                                        'version': version,
                                        'model_name' : model_name,
                                        'model_version' : model_version        
                                        }




# def add_predict_to_config(name='', func=None, hash='',model_name=None, model_version=None, requirements_file:str=None):
#     config = load_config()
#     code = ''
#     requirements = ''
    
#     if func is not None:
#         try:
#             code = get_source_code(func)
#             hash = generate_hash_for_function(func)

#             os.makedirs(PATH_PREDICT_DIR, exist_ok=True)
#             with open(PATH_PREDICT, 'w') as pred_file:
#                 pred_file.write(code)

#         except Exception as e:
#             print('Unable to get predict source code')
#             print(e)

#     if requirements_file is not None:
#         shutil.copy(requirements_file, PATH_PREDICT_REQUIREMENTS)


#         try:
#             with open(requirements_file, 'r') as req_file:
#                 requirements = req_file.readlines()
#                 requirements = [i.split('\n')[0] for i in requirements]
#         except Exception as e:
#             print('Unable to write requirements for prediction')
#             print(e)


#     config['predict'] = {
#                             'name' : name,
#                             'hash' : hash,
#                             'requirements': requirements,
#                             'code' : code,
#                             }

#     save_config(config=config)



def get_model_latest(config, version='latest'):
    config_model = config['model']
    model_name = None
    model_version = None
    model_hash = None

    model_positions = list(config_model.keys())


    if len(model_positions) != 0:
        # print(model_positions)
        position = model_positions[-1]
        model_name = config_model[position]['name']
        model_version = config_model[position]['version']
        model_hash = config_model[position]['hash']
    
    return model_name, model_version, model_hash



from pathlib import Path
from typing import Optional
import jwt
import requests
# import typer
from rich import print
from rich.syntax import Syntax

import os 
import json
import typing
from urllib.parse import urljoin

from . import get_token, get_project_id, get_org_id, convert_values_to_string
from pureml.utils.constants import BASE_URL, PATH_USER_PROJECT_DIR
from pureml.utils.pipeline import add_params_to_config


def post_params(params, model_name: str, model_version:str):
    user_token = get_token()
    org_id = get_org_id()
    project_id = get_project_id()
    
    url_path_1 = '{}/project/{}/model/{}/{}/params/add'.format(org_id, project_id, model_name, model_version)
    url = urljoin(BASE_URL, url_path_1)


    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': 'Bearer {}'.format(user_token)
    }

    params = json.dumps(params)
    data = {'model_name': model_name, 'params': params}

    response = requests.post(url, data=data, headers=headers)


    if response.status_code == 200:
        print(f"[bold green]Params have been registered!")

    else:
        print(f"[bold red]Params have not been registered!")

    return response


def add(params, model_name: str=None, model_version:str='latest') -> str:
    '''`add()` takes a dictionary of parameters and a model name as input and returns a string
    
    Parameters
    ----------
    params : dict
        a dictionary of parameters
    model_name : str
        The name of the model you want to add parameters to.
    model_version: str
        The version of the model
    
    Returns
    -------
        The response.text is being returned.
    
    '''

    params = convert_values_to_string(logged_dict=params)

    add_params_to_config(values=params, model_name=model_name, model_version=model_version)

    if model_name is not None and model_version is not None:
        response = post_params(params=params, model_name=model_name, model_version=model_version)

    #     return response.text
        
    # return 

        


# @app.command()
def fetch(model_name: str, model_version:str='latest', param:str='') -> str:
    '''
    
    This function fetches the parameters of a model
    
    Parameters
    ----------
    model_name : str
        The name of the model you want to fetch the parameters for.
    model_version: str
        The version of the model
    param : str
        The name of the parameter to fetch. If not specified, all parameters are returned.
    
    Returns
    -------
        The params that are fetched
    
    '''
    user_token = get_token()
    org_id = get_org_id()
    project_id = get_project_id()
    

    url_path_1 = '{}/project/{}/model/{}/{}/params/{}'.format(org_id, project_id, model_name, model_version, param)
    url = urljoin(BASE_URL, url_path_1)


    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': 'Bearer {}'.format(user_token)
    }


    response = requests.get(url, headers=headers)

    if response.status_code == 200:
        res_text = json.loads(response.text)

        if param == '':

            params = res_text

            # print(f"[bold green]Params have been fetched")
            # print(params)

            return params


        else:
            if 'param' in res_text.keys() and 'value' in res_text.keys():
                params = res_text['value']
                # params = json.loads(params)

                # print(f"[bold green]Params have been fetched")
                # print(res_text['param'], ':', res_text['value'])

                return params

            else:
                print('[bold red]Param {} are not available for the model!'.format(param))
                # print(response.text)
                return
        
            

    else:
        print(f"[bold red]Unable to fetch Params!")
        print(response.text)
        return


# @app.command()
def delete(param:str, model_name:str, model_version:str='latest') -> str:
    '''This function deletes a parameter from a model
    
    Parameters
    ----------
    model_name : str
        The name of the model you want to delete the parameter from.
    param : str
        The name of the parameter to delete.
    model_version: str
        The version of the model
    
    '''
    user_token = get_token()
    org_id = get_org_id()
    project_id = get_project_id()
    

    url_path_1 = '{}/project/{}/model/{}/{}/params/{}/delete'.format(org_id,project_id, model_name, model_version, param)
    url = urljoin(BASE_URL, url_path_1)


    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': 'Bearer {}'.format(user_token)
    }


    response = requests.delete(url, headers=headers)

    if response.status_code == 200:
        print(f"[bold green]Param has been deleted")
        
    else:
        print(f"[bold red]Unable to delete Param")

    return response.text



import typer
from pureml.config.parser import Config
# from pureml.trainer.train import Trainer
import os

app = typer.Typer()


import typer
from rich import print
# from puretrainer.train import Trainer
import os
from dotenv import load_dotenv

load_dotenv()


import pureml.cli.auth as auth
import pureml.cli.secrets as secrets

app = typer.Typer()
app.add_typer(auth.app, name="auth")
app.add_typer(secrets.app, name="secrets")


@app.callback(no_args_is_help=True)
def validate_user_authentication(ctx: typer.Context):
    # print(ctx.invoked_subcommand)
    if ctx.invoked_subcommand in ['auth']:
        return
    # user_token = auth.auth_validate()
    return



from pathlib import Path
from typing import Optional
import jwt
import requests
# import typer
from rich import print
from rich.syntax import Syntax

import os 
import json
import typing
from urllib.parse import urljoin

from . import get_token, get_project_id, get_org_id, convert_values_to_string
from pureml.utils.constants import BASE_URL, PATH_USER_PROJECT_DIR
from pureml.utils.pipeline import add_metrics_to_config


def post_metrics(metrics, model_name: str, model_version:str):
    user_token = get_token()
    org_id = get_org_id()
    project_id = get_project_id()
    
    url_path_1 = '{}/project/{}/model/{}/{}/metrics/add'.format(org_id,project_id, model_name, model_version)
    url = urljoin(BASE_URL, url_path_1)


    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': 'Bearer {}'.format(user_token)
    }

    metrics = json.dumps(metrics)
    data = {'model_name': model_name, 'metrics': metrics}

    response = requests.post(url, data=data, headers=headers)

    if response.status_code == 200:
        print(f"[bold green]Metrics have been registered!")
    
    else:
        print(f"[bold red]Metrics have not been registered!")

    return response

def add(metrics, model_name: str=None, model_version:str='latest') -> str:
    '''`add()` takes a dictionary of metrics and a model name as input and returns a string
    
    Parameters
    ----------
    metrics
        a dictionary of metrics
    model_name : str
        The name of the model you want to add metrics to.
    model_version: str
        The version of the model
    
    Returns
    -------
        The response.text is being returned.
    
    '''

    metrics = convert_values_to_string(logged_dict=metrics)

    add_metrics_to_config(values=metrics, model_name=model_name, model_version=model_version)

    if model_name is not None and model_version is not None:
        response = post_metrics(metrics=metrics, model_name=model_name, model_version=model_version)

        # return response.text
        
    # return 



def fetch(model_name: str, model_version:str='latest', metric:str='') -> str:
    '''This function fetches the metrics of a model
    
    Parameters
    ----------
    model_name : str
        The name of the model you want to fetch metrics for.
    model_version: str
        The version of the model
    metric : str
        The metric you want to fetch. If you want to fetch all the metrics, leave this parameter empty.
    
    Returns
    -------
        The metrics that are fetched
    
    '''
    user_token = get_token()
    org_id = get_org_id()
    project_id = get_project_id()
    

    url_path_1 = '{}/project/{}/model/{}/{}/metrics/{}'.format(org_id, project_id, model_name, model_version, metric)
    url = urljoin(BASE_URL, url_path_1)


    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': 'Bearer {}'.format(user_token)
    }


    response = requests.get(url, headers=headers)

    if response.status_code == 200:
        res_text = json.loads(response.text)

        if metric == '':

            metrics = res_text

            # print(f"[bold green]Metrics have been fetched")
            # print(metrics)

            return metrics


        else:
            if 'metric' in res_text.keys() and 'value' in res_text.keys():
                metrics = res_text['value']
                # metrics = json.loads(metrics)

                # print(f"[bold green]Metric has been fetched")
                # print(res_text['metric'], ':', res_text['value'])

                return metrics

            else:
                print('[bold red]Metric {} is not available for the model!'.format(metric))
                # print(response.text)
                return
        
            

    else:
        print(f"[bold red]Unable to fetch Metrics!")
        print(response.text)
        return



def delete(metric:str, model_name:str, model_version:str='latest') -> str:
    '''This function deletes a metric from a model
    
    Parameters
    ----------
    model_name : str
        The name of the model you want to delete the metric from
    metric : str
        The name of the metric to delete
    model_version: str
        The version of the model
    
    '''
    user_token = get_token()
    org_id = get_org_id()
    project_id = get_project_id()
    

    url_path_1 = '{}/project/{}/model/{}/{}/metrics/{}/delete'.format(org_id, project_id, model_name, model_version, metric)
    url = urljoin(BASE_URL, url_path_1)


    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': 'Bearer {}'.format(user_token)
    }


    response = requests.delete(url, headers=headers)

    if response.status_code == 200:
        print(f"[bold green]Metric has been deleted")
        
    else:
        print(f"[bold red]Unable to delete Metric")

    return response.text





import os


# BASE_URL = 'https://dev-api.pureml.com'
BASE_URL = 'https://api.pureml.com'

PATH_USER_TOKEN =  os.path.join(os.path.expanduser('~'), '.pureml/token')

PATH_PUREML_RELATIVE = '.pureml'

PATH_USER_PROJECT_DIR = os.path.join(os.getcwd(),PATH_PUREML_RELATIVE)

PATH_USER_PROJECT = os.path.join(PATH_USER_PROJECT_DIR,'pure.project')


PATH_CONFIG = os.path.join(PATH_USER_PROJECT_DIR,'config.pkl')# 'temp.yaml'

PATH_ARTIFACT_DIR = os.path.join(PATH_USER_PROJECT_DIR,'artifacts')
PATH_ARRAY_DIR = os.path.join(PATH_USER_PROJECT_DIR,'array')
PATH_AUDIO_DIR = os.path.join(PATH_USER_PROJECT_DIR,'audio')
PATH_FIGURE_DIR = os.path.join(PATH_USER_PROJECT_DIR,'figure')
PATH_TABULAR_DIR = os.path.join(PATH_USER_PROJECT_DIR,'tabular')
PATH_VIDEO_DIR = os.path.join(PATH_USER_PROJECT_DIR,'video')
PATH_IMAGE_DIR = os.path.join(PATH_USER_PROJECT_DIR,'image')

PATH_DATASET_DIR = os.path.join(PATH_USER_PROJECT_DIR,'dataset')
PATH_MODEL_DIR = os.path.join(PATH_USER_PROJECT_DIR,'model')


PATH_PREDICT_DIR_RELATIVE = 'predict'
PATH_PREDICT_DIR = os.path.join(PATH_USER_PROJECT_DIR,PATH_PREDICT_DIR_RELATIVE)
PATH_PREDICT_REQUIREMENTS = os.path.join(PATH_PREDICT_DIR,'requirements.txt')
PATH_PREDICT = os.path.join(PATH_PREDICT_DIR,'predict.py')

PATH_PREDICT_USER = os.path.join(os.getcwd(), 'predict.py')
PATH_PREDICT_REQUIREMENTS_USER = os.path.join(os.getcwd(), 'requirements.txt')


PATH_FASTAPI_FILE = os.path.join(PATH_PREDICT_DIR,'fastapi_server.py')
PORT_FASTAPI = 8005      #Same port as docker server


# PATH_DOCKER_DIR = os.path.join(PATH_USER_PROJECT_DIR,'docker')
PATH_DOCKER_IMAGE = os.path.join(PATH_PREDICT_DIR,'Dockerfile')
PATH_DOCKER_CONFIG = os.path.join(PATH_PREDICT_DIR,'DockerConfig.yaml')
# PORT_DOCKER = 8005      #Same port as fastapi server
PORT_HOST = 8000
BASE_IMAGE_DOCKER = 'python:3.8-slim'
API_IP_DOCKER = "0.0.0.0"
API_IP_HOST = "0.0.0.0"





import os
import joblib
from collections import defaultdict, OrderedDict
from .constants import PATH_CONFIG


def load_config():
    if os.path.exists(PATH_CONFIG):
        config = joblib.load(PATH_CONFIG)
    else:
        config = defaultdict()
        
        config['load_data'] = defaultdict()
        config['transformer'] = OrderedDict()
        config['dataset'] = defaultdict()
        
        config['model'] = OrderedDict()
        
        config['params'] = defaultdict()
        config['metrics'] = defaultdict()
        config['artifacts'] = defaultdict()

        config['predict'] = defaultdict()
        
        joblib.dump(config, PATH_CONFIG)

    return config

def save_config(config):
    save_dir = os.path.dirname(PATH_CONFIG)
    
    os.makedirs(save_dir, exist_ok=True)
    # print('config')
    # print(config)
    
    joblib.dump(config, PATH_CONFIG)





from pathlib import Path
from typing import Optional
import jwt
import requests
# import typer
from rich import print
from rich.syntax import Syntax

import os 
import json
import typing

from urllib.parse import urljoin

from . import get_token, get_project_id, get_org_id

from pureml.utils.constants import BASE_URL, PATH_TABULAR_DIR
from joblib import Parallel, delayed


def details(model_name:str, model_version:str='latest', name:str=''):
    '''This function returns the details of the tabular for a given model
    
    Parameters
    ----------
    model_name : str
        The name of the model you want to get the tabular details for
    model_version: str
        The version of the model
    name : str
        The name of the tabular.
    
    '''

    user_token = get_token()
    org_id = get_org_id()
    project_id = get_project_id()



    url_path_1 = '{}/project/{}/model/{}/{}/tabular/{}/'.format(org_id, project_id, model_name, model_version, name)
    url = urljoin(BASE_URL, url_path_1)

    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': 'Bearer {}'.format(user_token)
    }
    

    response = requests.get(url, headers=headers)


    if response.status_code == 200:
        res_text = json.loads(response.text)

        if len(res_text) == 0:
            print('[bold yellow] No Tabular have been found for the model')
            print(res_text)
            return 
        else:
            print('[bold green]Tabular have been found for the model')
            print(res_text)
            return res_text

    else:
        print('[bold red]Unable to obtain the tabular details')
        print(response.text)
        return




def add(tabular: str, model_name: str, model_version:str='latest') -> str:    
    '''`add` function takes in the path of the tabular, name of the tabular and the model name and
    registers the tabular
    
    Parameters
    ----------
    tabular : str
        The path to the tabular file.
    name : str
        The name of the tabular.
    model_name : str
        The name of the model you want to add Tabular to.
    model_version: str
        The version of the model
    
    Returns
    -------
        The response is a JSON object
    
    '''
    
    user_token = get_token()
    org_id = get_org_id()
    project_id = get_project_id()
    
    url_path_1 = '{}/project/{}/model/{}/{}/tabular/add'.format(org_id, project_id, model_name, model_version)
    url = urljoin(BASE_URL, url_path_1)


    user_token = get_token()
    project_id = get_project_id()

    headers = {
        'Authorization': 'Bearer {}'.format(user_token)
    }

    files = {}
    for file_name, file_path in tabular.items():
        
        if os.path.isfile(file_path):
            files[file_name] = open(file_path, 'rb')
        else:
            print('[bold red] Tabular', file_name,'doesnot exist at the given path')

    
    data = {'name_path_mapping' : tabular}

    response = requests.post(url, data=data, files=files, headers=headers)
    
    
    if response.status_code == 200:
        print(f"[bold green]tabulars have been registered!")

    else:
        print(f"[bold red]tabulars have not been registered!")
        print(response.text)

    return response.text


def fetch(model_name: str, model_version:str='latest', name:str = ''):
    '''It fetches the tabular from the server and stores it in the local directory
    
    Parameters
    ----------
    model_name : str
        The name of the model you want to fetch the tabular from.
    model_version: str
        The version of the model
    name : str
        The name of the tabular to be fetched. If not specified, all tabulars will be fetched.
    
    Returns
    -------
        The response text is being returned.
    
    '''

    user_token = get_token()
    org_id = get_org_id()
    project_id = get_project_id()


    def fetch_tabular(tabular_details: dict):

        url = tabular_details['location']
        file_path_temp = tabular_details['path']
        file_name = file_path_temp.split(os.path.sep)[-1]
        save_path = os.path.join(PATH_TABULAR_DIR, file_name)
        print('save path', save_path)

        name_fetched = tabular_details['tabular']


        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'Authorization': 'Bearer {}'.format(user_token)
        }
        
        print('tabular url', url)

        # response = requests.get(url, headers=headers)
        response = requests.get(url)

        print(response.status_code)

        if response.status_code == 200:
            print('[bold green] tabular {} has been fetched'.format(name_fetched))

            save_dir = os.path.dirname(save_path)

            os.makedirs(save_dir, exist_ok=True)

            tabular_bytes = response.content

            open(save_path, 'wb').write(tabular_bytes)


            print('[bold green] tabular {} has been stored at {}'.format(name_fetched, save_path))
            
            return response.text
        else:
            print('[bold red] Unable to fetch the tabular')

            return response.text


    tabular_details = details(model_name=model_name, name=name, model_version=model_version)

    if tabular_details is None:
        return

    if type(tabular_details) is dict:

        res_text = fetch_tabular(tabular_details)

    elif type(tabular_details) is list:
        res_text = Parallel(n_jobs=-1)(delayed(fetch_tabular)(art_det) for art_det in tabular_details)


    return res_text
    


def delete(name:str, model_name:str,  model_version:str='latest') -> str:
    '''`delete()` deletes an tabular from a model
    
    Parameters
    ----------
    name : str
        The name of the tabular you want to delete.
    model_name : str
        The name of the model you want to delete the tabular from
    model_version: str
        The version of the model
    
    '''

    user_token = get_token()
    org_id = get_org_id()
    project_id = get_project_id()

    url_path_1 = '{}/project/{}/model/{}/{}/tabular/{}/delete'.format(org_id, project_id, model_name, model_version, name)
    url = urljoin(BASE_URL, url_path_1)

    
    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': 'Bearer {}'.format(user_token)
    }
    

    # tabular_details = details(model_name=model_name, tabular=tabular)

    # if tabular_details is None:
    #     print('[bold red] Unable to find tabular details')
    #     return


    response = requests.delete(url, headers=headers)


    if response.status_code == 200:
        print(f"[bold green]tabular has been deleted")
        
    else:
        print(f"[bold red]Unable to delete tabular")

    return response.text






from pureml.utils.constants import BASE_URL
from urllib.parse import urljoin
import requests
import json
from rich import print
from pureml.cli.auth import save_auth
from . import get_org_id, get_token

def login(org_id:str, access_token:str) -> str:
    ''' The function takes in a user API token and logs in a user for a session.
    
    Parameters
    ----------
    token: str
        API token for the user. This token will be used to authenticate an user.
    
    '''
    
    url_path_1 = 'org/id/{}'.format(org_id)
    url = urljoin(BASE_URL, url_path_1)


    headers = {
        'Authorization': 'Bearer {}'.format(access_token)
    }


    response = requests.get(url, headers=headers)

    # print(response.text)
    if response.status_code == 200:

        response_text = response.text
        response_org_details = json.loads(response_text)['data']
        
        # if response_org_details is not None:
        response_org_id = response_org_details[0]['id']

        if response_org_id == org_id:
            print('[green] Valid Org Id and Access token')
            save_auth(org_id=org_id, access_token=access_token)
        
        else:
            print('[orange] Valid Org Id and Access token. Obtained different organization')

        # else:
        #     print('[green] Invalid Org Id and Access token')
    elif response.status_code == 403:
        print('[red] Invalid Access token')
    elif response.status_code == 404:
        print('[red] Invalid Org Id')
    else:
        print("[red] Unable to obtain the organization details")



    
def details():
    token = get_token()
    org_id = get_org_id()

    print('Org Id: ', org_id)
    print('Access Token: ', token)        




import os, json
from pureml.utils.constants import PATH_USER_TOKEN, PATH_USER_PROJECT





def get_token():
    '''It checks if the token exists in the user's home directory. If it does, it returns the token. If it
    doesn't, it returns None
    
    Returns
    -------
        The token is being returned.    
    '''
    path = PATH_USER_TOKEN
    # path = os.path.expanduser(path)

    if os.path.exists(path):
        creds = open(path, "r").read()

        creds_json = json.loads(creds)
        token = creds_json['accessToken']
        # print(f"[bold green]Authentication token exists!")

        # print(token)
        return token
    else:
        print(f"[bold red]Authentication token doesnot exist! Please login")

        return
    
    

def get_org_id():
    '''It checks if the org exists in the user's home directory. If it does, it returns the org. If it
    doesn't, it returns None
    
    Returns
    -------
        The org is being returned.
    
    '''
    path = PATH_USER_TOKEN

    path = os.path.expanduser(path)

    if os.path.exists(path):
        creds = open(path, "r").read()

        creds_json = json.loads(creds)

        org_id = creds_json['email']
        # print(f"[bold green]Organization exists!")

        # print(org_id)
        return org_id
    else:
        print(f"[bold red]Organization token doesnot exist! Please login")

        return



def get_project_id():
    '''It checks if the project file exists in the current directory. If it does, it reads the project file
    and returns the project id
    
    Returns
    -------
        The project id is being returned.
    
    '''
    project_file_name = PATH_USER_PROJECT
    project_path = os.path.join(os.getcwd(), project_file_name)

    if os.path.exists(project_path):
        project_details = open(project_path, "r").read()
        # print(project_details, type(project_details))

        project_details = json.loads(project_details)
        # project_details = project_details['data'][0]

        # print(project_details, type(project_details))
        project_id = project_details['id']


        # print('Project exists locally. Project Id:', project_id)
        # print(project_id)
        return project_id
    
    return

def get_project_name():
    '''It checks if the project file exists in the current directory. If it does, it reads the project file
    and returns the project id
    
    Returns
    -------
        The project id is being returned.
    
    '''
    # project_file_name = PATH_USER_PROJECT
    # project_path = os.path.join(os.getcwd(), project_file_name)
    project_path = PATH_USER_PROJECT
    print(project_path)

    if os.path.exists(project_path):
        project_details = open(project_path, "r").read()
        # print(project_details, type(project_details))

        project_details = json.loads(project_details)
        # project_details = project_details['data'][0]

        # print(project_details, type(project_details))
        project_name = project_details['name']


        # print('Project exists locally. Project Id:', project_id)
        # print(project_id)
        return project_name
    
    return



def convert_values_to_string(logged_dict):
    
    for key in logged_dict:
        logged_dict[key] = str(logged_dict[key])

    return logged_dict



from collections import defaultdict
import hashlib
import os
import joblib
from .constants import PATH_CONFIG
from .config import load_config, save_config
import json
import inspect
import re


def file_reader_chunk(file_obj, chunk_size=1024):
    """Generator that reads a file in chunks of bytes"""
    while True:
        chunk = file_obj.read(chunk_size)
        if not chunk:
            return
        yield chunk


def generate_hash_file(file_path, hash=hashlib.md5):
    hash_obj = hash()
    file_object = open(file_path, 'rb')

    
    for chunk in file_reader_chunk(file_object):
        hash_obj.update(chunk)
    
    hash_value = hash_obj.hexdigest()

    file_object.close()
    
    return hash_value





def generate_hash_for_dict(values, hash=hashlib.md5):
    value_str = json.dumps(values).encode()
    
    file_object = hash(value_str)

    hash_value = file_object.hexdigest()
    
    return hash_value



def generate_hash_for_function(func, hash=hashlib.md5):

    string_stripped = re.sub(r"[\n\t\s]*", "", inspect.getsource(func))

    string_stripped = string_stripped.encode()
    
    string_object = hash(string_stripped)

    hash_value = string_object.hexdigest()
    
    return hash_value




def check_hash_status_model(file_path, name, item_key='model'):
    hash_value = generate_hash_file(file_path=file_path)
    hash_status = False

    config = load_config()
    if len(config) > 0:
        models = config[item_key]
        for model_invoke_order, model_details in models.items():
            # print('model_invoke_order', model_invoke_order)
            # print('model_details', model_details)

            if hash_value == model_details['hash']:
                hash_status = True
                return hash_status, hash_value

            
    return hash_status, hash_value


def check_hash_status_dataset(file_path, name, item_key='dataset'):
    hash_value = generate_hash_file(file_path=file_path)
    hash_status = False

    config = load_config()
    if len(config) > 0:
        dataset = config[item_key]


        if hash_value == dataset['hash']:
            hash_status = True
            return hash_status, hash_value

            
    return hash_status, hash_value






import os

from .utils.constants import PATH_USER_PROJECT_DIR
os.makedirs(PATH_USER_PROJECT_DIR, exist_ok=True)


from .packaging import load_model, save_model

from .components import project as project
from .components import model as model
from .components import dataset as dataset
from .components import metrics as metrics
from .components import params as params
from .components import artifacts as artifacts
from .components import auth as auth
from .components.auth import login
from .components.log import log

from .deploy import docker


from pureml.utils.constants import PATH_CONFIG
from pureml.utils.config import load_config
from collections import OrderedDict
import json


def create_nodes(components):
    # print(components)

    nodes = [{'id': component['name'], 'text': component['name']} for component in components]


    return nodes


def create_extra_nodes(nodes, edges):
    # print(components)
    
    node_nodes = [i['id'] for i in nodes]
    edge_nodes = sum([[i['from']]+[i['to']] for i in edges], [])

    extra_nodes = list(set([n for n in edge_nodes if n not in node_nodes]))

    nodes = nodes + [{'id': n, 'text':n} for n in extra_nodes]


    return nodes




def create_edges(components):
    
    edges = []

    def add_node(node_from, node_to):
        edge_id = '--->'.join([node_from, node_to])
        edge = {
            'id': edge_id,
            'from': node_from,
            'to': node_to            
        }

        edges.append(edge)


    for n in components:
        if 'parent' in n.keys():
            node_parent = n['parent']
            node_to = n['name']
                    

            if node_parent is not None:

                if type(node_parent) == list:
                
                    for node_from in node_parent:
                        add_node(node_from, node_to)

                else:
                    add_node(node_parent, node_to)



    return edges


def prune_dict_elements(components):
    components_string = [json.dumps(d) for d in components]

    components_string = set(components_string)
    
    components_loaded = [json.loads(d) for d in components_string]

    return components_loaded




def create_pipeline():
    config = load_config()

    load_data = config['load_data']
    transformer = list(config['transformer'].values())
    dataset = config['dataset']



    pipeline_components = []

    if len(load_data) > 0:
        pipeline_components.append(load_data)
    
    if len(transformer) > 0:
        pipeline_components = pipeline_components + transformer
    
    if len(dataset) > 0:
        pipeline_components.append(dataset)


    pipeline_components =  prune_dict_elements(pipeline_components)

    edges = create_edges(components=pipeline_components)
    edges = prune_dict_elements(edges)


    nodes = create_nodes(components=pipeline_components)
    nodes = create_extra_nodes(nodes, edges)
    nodes = prune_dict_elements(nodes)


    pipeline = {
        'edges': edges,
        'nodes': nodes
    }

    return pipeline





    




# const nodes = [
#   {
#     id: '1',
#     text: '1'
#   },
#   {
#     id: '2',
#     text: '2'
#   }
# ];

# const edges = [
#   {
#     id: '1-2',
#     from: '1',
#     to: '2'
#   }
# ];




import typing 
import numpy as np
import matplotlib.pyplot as plt
from pureml.components import metrics as pure_metrics
from pureml.components import params as pure_params


# def log(**kwargs):

#     if 'metrics' in kwargs.keys():
#         metrics.add(kwargs['metrics'])

    
#     if 'params' in kwargs.keys():
#         params.add(kwargs['params'])

def log(metrics = None, params=None, model_name=None, model_version=None):


    if metrics is not None:
        func_params = {}

        if model_name is not None:
            func_params['model_name'] = model_name
        if model_version is not None:
            func_params['model_version'] = model_version

        func_params['metrics'] = metrics.copy()
        
        pure_metrics.add(**func_params)

    if params is not None:
        func_params = {}

        if model_name is not None:
            func_params['model_name'] = model_name
        if model_version is not None:
            func_params['model_version'] = model_version

        func_params['params'] = params.copy()
        pure_params.add(**func_params)



from pydantic import BaseModel, root_validator

from abc import ABC, abstractmethod

import typing
import pickle as pkl

from pydantic import BaseModel
import typing
import joblib

from pureml.packaging.errors import FrameworkNotSupportedError
from .model_framework import ModelFramework, ModelFrameworkType
# from . import MODEL_FRAMEWORKS_BY_TYPE, SUPPORTED_MODEL_FRAMEWORKS


from .packaging_utils import get_file_size
from .model_packaging.sklearn import SKLearn
from .model_packaging.catboost import CatBoost
from .model_packaging.xgboost import XGBoost
from .model_packaging.lightgbm import LightGBM
from .model_packaging.keras import Keras
from .model_packaging.tensorflow import Tensorflow
from .model_packaging.pytorch import Pytorch
from .model_packaging.pytorch_tabnet import PytorchTabnet
from .model_packaging.custom import Custom
from pureml.utils.constants import PATH_MODEL_DIR




MODEL_FRAMEWORKS_BY_TYPE = {
    ModelFrameworkType.SKLEARN: SKLearn(),
    ModelFrameworkType.XGBOOST: XGBoost(),
    ModelFrameworkType.LIGHTGBM: LightGBM(),
    ModelFrameworkType.CATBOOST: CatBoost(),
    ModelFrameworkType.KERAS: Keras(),
    # ModelFrameworkType.HUGGINGFACE_TRANSFORMER: HuggingfaceTransformer(),
    ModelFrameworkType.PYTORCH: Pytorch(),
    ModelFrameworkType.PYTORCH_TABNET: PytorchTabnet(),
    ModelFrameworkType.CUSTOM: Custom()
}


SUPPORTED_MODEL_FRAMEWORKS = [
    ModelFrameworkType.SKLEARN,
    ModelFrameworkType.XGBOOST,
    ModelFrameworkType.LIGHTGBM,
    ModelFrameworkType.KERAS,
    ModelFrameworkType.TENSORFLOW,
    # ModelFrameworkType.HUGGINGFACE_TRANSFORMER,
    ModelFrameworkType.PYTORCH,
    ModelFrameworkType.PYTORCH_TABNET,
    ModelFrameworkType.CUSTOM
]


class Model(ABC, BaseModel):
    model: typing.Any = None
    model_name: str = 'model'
    model_path: str = None
    model_class: str = None

    model_config: dict = None
    model: typing.Any = None
    model_framework: typing.Any = None
    model_requirements: list = None

    #By default predict function of a framework should be assigned to here
    #If a user gives a predict function, assign it here
    predict: typing.Any = None



    @root_validator
    def set_fields(cls, values):


        return values



    # @staticmethod
    def from_dict(self):
        
        self.model = self.model_config['model']
        # model_name = model_config_dict['model_name'],
        self.model_framework = self.model_config['model_framework']
        self.model_requirements = self.model_config['model_requirements']





    def generate_model_config(self):

        model_config = {
                            'model': self.model,
                            'model_framework': self.model_framework,
                            'model_requirements': self.model_requirements,
                            'model_size': get_file_size(pkl.dumps(self.model))
                        }

        return model_config



    def model_framework_from_model(self) -> ModelFramework:
        self.model_class = self.model.__class__
        model_framework = self.model_framework_from_model_class(self.model_class)

        return model_framework




    def model_framework_from_model_class(self, model_class) -> ModelFramework:

        for framework in MODEL_FRAMEWORKS_BY_TYPE.values():

            if framework.supports_model_class(model_class):
                return framework

        # raise FrameworkNotSupportedError(
        #     "Model must be one of "
        #     + "/".join([t.value for t in SUPPORTED_MODEL_FRAMEWORKS])
        # )
        framework = ModelFrameworkType.CUSTOM
        return framework



    def save_model(self):
        self.model_framework = self.model_framework_from_model()
        self.model_requirements = self.model_framework.get_requirements()
 
        # self.model_framework = ''
        # self.model_requirements = []
        
        self.model_config = self.generate_model_config()
        
        joblib.dump(self.model_config, self.model_path)


        return self.model_path


    def load_model(self):
        self.model_config = joblib.load(self.model_path)
        self.from_dict() 


        return self.model















from pureml.utils.constants import PATH_PREDICT_DIR_RELATIVE, PATH_DOCKER_IMAGE, PATH_DOCKER_CONFIG, API_IP_HOST
from pureml.utils.constants import PORT_FASTAPI, PORT_HOST, BASE_IMAGE_DOCKER, PATH_FASTAPI_FILE, PATH_PREDICT_REQUIREMENTS, PATH_PREDICT_DIR
import os
import docker
from .fastapi import create_fastapi_file




def create_docker_file(org_id, access_token):
    # os.makedirs(PATH_DOCKER_DIR, exist_ok=True)
    os.makedirs(PATH_PREDICT_DIR, exist_ok=True)

    req_pos = PATH_PREDICT_REQUIREMENTS.find(PATH_PREDICT_DIR_RELATIVE)
    req_path = PATH_PREDICT_REQUIREMENTS[req_pos:]

    api_pos = PATH_FASTAPI_FILE.find(PATH_PREDICT_DIR_RELATIVE)
    api_path = PATH_FASTAPI_FILE[api_pos:]
 

    docker = """
    
FROM {BASE_IMAGE}

ENV ORG_ID={ORG_ID}
ENV ACCESS_TOKEN={ACCESS_TOKEN}

RUN mkdir -p {PREDICT_DIR}

WORKDIR {PREDICT_DIR}

ADD . {PREDICT_DIR}

COPY . .

RUN pip install --no-cache-dir --upgrade pip \
  && pip install --no-cache-dir -r {REQUIREMENTS_PATH}

RUN pip install pureml fastapi uvicorn python-multipart

EXPOSE {PORT}
CMD ["python", "{API_PATH}"]    
""".format(
        BASE_IMAGE=BASE_IMAGE_DOCKER,
        PORT=PORT_FASTAPI ,
        PREDICT_DIR = PATH_PREDICT_DIR_RELATIVE,
        API_PATH=api_path,
        REQUIREMENTS_PATH=req_path,
        ORG_ID = org_id,
        ACCESS_TOKEN = access_token
    )


    with open(PATH_DOCKER_IMAGE, "w") as docker_file:
        docker_file.write(docker)



    docker_yaml = """version: '3'

services:
  prediction:
    build: .
    container_name: "{CONTAINER_NAME}"
    expose:
      - "{DOCKER_PORT}"
    ports:
      - "{HOST_PORT}:{DOCKER_PORT}"
    
    """.format(DOCKER_PORT=PORT_FASTAPI, HOST_PORT=PORT_HOST,
               CONTAINER_NAME='pureml_prediction')
    
    
    with open(PATH_DOCKER_CONFIG, "w") as docker_yaml_file:
        docker_yaml_file.write(docker_yaml)

    



def create_docker_image(image_tag=None):
  if image_tag is None:
    image_tag = 'pureml_docker_image'
  else:
    image_tag = image_tag.replace(' ', '')

    client = docker.from_env()

    docker_file_path_relative = PATH_DOCKER_IMAGE.split(os.path.sep)[-1]

    try:
      image, build_log  = client.images.build(path=PATH_PREDICT_DIR, 
                                              dockerfile=docker_file_path_relative,
                                              tag=image_tag, nocache=True,
                                              rm=True)
      
      print('Docker image is created')
      print(image)

    except Exception as e:
      print(e)
      image = None

    return image




def run_docker_container(image, name):
  client = docker.from_env()

  docker_port = '{port}/tcp'.format(port=PORT_FASTAPI)
  # print(docker_port)

  container = client.containers.run(image=image, ports={docker_port: PORT_HOST}, 
                                    detach=True, name=name)

  return container


def create(model_name, model_version, org_id, access_token, image_tag=None, 
            predict_path=None, requirements_path=None, container_name=None,
            input:dict=None, output:dict=None):

  create_fastapi_file(model_name=model_name, model_version=model_version,
                      predict_path=predict_path, requirements_path=requirements_path,
                      input=input, output=output)

  create_docker_file(org_id=org_id, access_token=access_token)

  image = create_docker_image(image_tag)

  if image is not None:
    container = run_docker_container(image=image, name=container_name)

    print('Created Docker container')
    print(container)
    print('Prediction requests can be forwarded to {ip}:{port}/predict'.format(ip=API_IP_HOST, port=PORT_HOST))
  else:
    print('Failed to create the container')


  
def get(container_id):

  client = docker.from_env()
  
  container = client.containers.get(container_id)

  return container



def stop(container_id):

  client = docker.from_env()

  container = client.containers.get(container_id)

  container.stop()



import os
import shutil

from pureml.utils.constants import (
    API_IP_DOCKER,
    PATH_FASTAPI_FILE,
    PATH_PREDICT,
    PATH_PREDICT_DIR,
    PATH_PREDICT_REQUIREMENTS,
    PATH_PREDICT_REQUIREMENTS_USER,
    PATH_PREDICT_USER,
    PATH_USER_PROJECT,
    PORT_FASTAPI,
)
from pureml.utils.deploy import process_input, process_output


def get_project_file():
    os.makedirs(PATH_PREDICT_DIR, exist_ok=True)

    project_dir_name = PATH_USER_PROJECT.split(os.path.sep)[-2]
    predict_project_dir = os.path.join(PATH_PREDICT_DIR, project_dir_name)

    os.makedirs(predict_project_dir, exist_ok=True)

    project_file_name = PATH_USER_PROJECT.split(os.path.sep)[-1]
    predict_project_file_name = os.path.join(predict_project_dir, project_file_name)

    shutil.copy(PATH_USER_PROJECT, predict_project_file_name)


def get_predict_file(predict_path):

    os.makedirs(PATH_PREDICT_DIR, exist_ok=True)

    if predict_path is None:
        predict_path = PATH_PREDICT_USER
        print("Taking the default predict.py file path: ", predict_path)
    else:
        print("Taking the predict.py file path: ", predict_path)

    if os.path.exists(predict_path):
        shutil.copy(predict_path, PATH_PREDICT)
    else:
        raise Exception(predict_path, "doesnot exists!!!")


def get_requirements_file(requirements_path):

    os.makedirs(PATH_PREDICT_DIR, exist_ok=True)

    if requirements_path is None:
        requirements_path = PATH_PREDICT_REQUIREMENTS_USER
        print("Taking the default requirements.txt file path: ", requirements_path)
    else:
        print("Taking the requirements.txt file path: ", requirements_path)

    if os.path.exists(requirements_path):
        shutil.copy(requirements_path, PATH_PREDICT_REQUIREMENTS)
    else:
        raise Exception(requirements_path, "doesnot exists!!!")


def generate_file_upload_api(
    model_name, model_version, input_type, input_shape, output_type, output_shape
):

    query = """
from fastapi import FastAPI, Depends, Request, UploadFile
import uvicorn
import pureml
from predict import model_predict
import os
from dotenv import load_dotenv
import json
import shutil
from pureml.utils.deploy import parse_input, parse_output
from typing import Union

load_dotenv()

org_id = os.getenv('ORG_ID')
access_token = os.getenv('ACCESS_TOKEN')

pureml.login(org_id=org_id, access_token=access_token)

model = pureml.model.fetch('{MODEL_NAME}', '{MODEL_VERSION}')

# Create the app
app = FastAPI()     

@app.post('/predict')
async def predict(file: Union[UploadFile, None] = None):
    input_type = '{INPUT_TYPE}'
    input_shape = '{INPUT_SHAPE}'
    output_type = '{OUTPUT_TYPE}'
    output_shape = '{OUTPUT_SHAPE}'

    if input_type == 'None':
        print('Rebuild the docker container with non null input_type')
        predictions = json.dumps(None)
        return predictions

    path = None

    if not file:
        print('No upload file sent')
        predictions = json.dumps(None)
        return predictions
    else:
        path = file.filename
        print(path)
        with open(path, "wb") as buffer:
            shutil.copyfileobj(file.file, buffer)

    if path is None:
        print('Input imagepath is None')
        predictions = json.dumps(None)
        return predictions


    data = parse_input(data=path, input_type=input_type, input_shape=input_shape)

    if data is None:
        print('Error in data input format')
        predictions = json.dumps(None)
        return predictions


    predictions = model_predict(model, data)

    predictions = parse_output(data=predictions, output_type=output_type, output_shape=output_shape)


    return predictions


if __name__ == '__main__':
    uvicorn.run(app, host='{HOST}', port={PORT})""".format(
        HOST=API_IP_DOCKER,
        PORT=PORT_FASTAPI,
        MODEL_NAME=model_name,
        MODEL_VERSION=model_version,
        INPUT_TYPE=input_type,
        INPUT_SHAPE=input_shape,
        OUTPUT_TYPE=output_type,
        OUTPUT_SHAPE=output_shape,
    )

    return query


def generate_json_api(
    model_name, model_version, input_type, input_shape, output_type, output_shape
):

    query = """
from fastapi import FastAPI, Depends, Request
import uvicorn
import pureml
from predict import model_predict
import os
from dotenv import load_dotenv
import pandas as pd
import json
import numpy as np
from pureml.utils.deploy import parse_input, parse_output

load_dotenv()

org_id = os.getenv('ORG_ID')
access_token = os.getenv('ACCESS_TOKEN')

pureml.login(org_id=org_id, access_token=access_token)

model = pureml.model.fetch('{MODEL_NAME}', '{MODEL_VERSION}')

# Create the app
app = FastAPI()     

@app.post('/predict')
async def predict(request: Request):
    input_type = '{INPUT_TYPE}'
    input_shape = '{INPUT_SHAPE}'
    output_type = '{OUTPUT_TYPE}'
    output_shape = '{OUTPUT_SHAPE}'

    if input_type == 'None':
        print('Rebuild the docker container with non null input_type')
        predictions = json.dumps(None)
        return predictions

    req_json = await request.json()

    data_json = req_json['test_data']


    data = parse_input(data=data_json, input_type=input_type, input_shape=input_shape)

    if data is None:
        print('Error in data input format')
        predictions = json.dumps(None)
        return predictions


    predictions = model_predict(model, data)

    predictions = parse_output(data=predictions, output_type=output_type, output_shape=output_shape)


    return predictions


if __name__ == '__main__':
    uvicorn.run(app, host='{HOST}', port={PORT})""".format(
        HOST=API_IP_DOCKER,
        PORT=PORT_FASTAPI,
        MODEL_NAME=model_name,
        MODEL_VERSION=model_version,
        INPUT_TYPE=input_type,
        INPUT_SHAPE=input_shape,
        OUTPUT_TYPE=output_type,
        OUTPUT_SHAPE=output_shape,
    )

    return query


def generate_api(input, output, model_name, model_version):

    input_type, input_shape = process_input(input=input)
    output_type, output_shape = process_output(output=output)

    if input_type == "image":
        api = generate_file_upload_api(
            model_name=model_name,
            model_version=model_version,
            input_type=input_type,
            input_shape=input_shape,
            output_type=output_type,
            output_shape=output_shape,
        )
    else:
        api = generate_json_api(
            model_name=model_name,
            model_version=model_version,
            input_type=input_type,
            input_shape=input_shape,
            output_type=output_type,
            output_shape=output_shape,
        )

    return api


def create_fastapi_file(
    model_name, model_version, predict_path, requirements_path, input, output
):

    get_project_file()

    get_predict_file(predict_path)

    get_requirements_file(requirements_path)

    api = generate_api(
        input=input, output=output, model_name=model_name, model_version=model_version
    )

    with open(PATH_FASTAPI_FILE, "w") as api_writer:
        api_writer.write(api)

    api_writer.close()

    print("FastAPI server files are created")


#     print("""
#           API sucessfully created. To run your API, please run the following command
# --> !python <api_name>
#           """)


def run_fastapi_server():
    pass



import os
from pathlib import Path
from typing import Optional
import jwt
import requests
import typer
from rich import print
from rich.syntax import Syntax
from pureml.utils.constants import BASE_URL, PATH_USER_TOKEN
from pureml.components import get_org_id, get_token

from urllib.parse import urljoin
import json

app = typer.Typer()


def save_auth(org_id:str=None, access_token:str=None):
    token_path = PATH_USER_TOKEN

    token_dir = os.path.dirname(token_path)
    os.makedirs(token_dir, exist_ok=True)


    token = {
                'email': org_id,
                'accessToken': access_token
            }
        
    token = json.dumps(token)
    
    with open(token_path, "w") as token_file:
        token_file.write(token)

    print('[green]User token is stored')



@app.command()
def details():
    token = get_token()
    org_id = get_org_id()

    print('Org Id: ', org_id)
    print('Access Token: ', token)






@app.callback()
def callback():
    """
    Authentication user command

    Use with status, signup, login or logout option

    status - Checks current authenticated user
    signup - Creates new user
    login - Logs in user
    logout - Logs out user
    """

@app.command()
def signup():
    print("\n[bold]Create a new account[/bold]\n")
    email: str = typer.prompt("Enter new email")
    password: str = typer.prompt("Enter new password", confirmation_prompt=True, hide_input=True)
    # organization_id: str = typer.prompt("Enter Organization id")
    # data = {"email": email, "password": password, "org": organization_id}
    data = {"email": email, "password": password}


    url_path_1 = 'user/signup'
    url = urljoin(BASE_URL, url_path_1)

    response = requests.post(url,json=data)


    if not response.ok:
        print(f"[bold red]Could not create account! Please try again later")
        return
    print(f"[bold green]Successfully created your account! You can now login using ```pure auth login```")

@app.command()
def login():
    print(f"\n[bold]Enter your credentials to login[/bold]\n")
    email: str = typer.prompt("Enter your email")
    password: str = typer.prompt("Enter your password", hide_input=True)
    data = {"email": email, "password": password}


    url_path_1 = 'user/login'
    url = urljoin(BASE_URL, url_path_1)

    response = requests.post(url,json=data)
    
    if not response.ok:
        print(f"[bold red]Could not login! Please try again later")
        print('[bold yellow] Please signup at https://app.pureml.com/auth/signup')
        return
    elif response == "":
        print(f"[bold red]Invalid email or password!")
        print('[bold yellow] Please signup at https://app.pureml.com/auth/signup')
        return


    if response.status_code == 200:
        token = response.text
        token = json.loads(token)['data'][0]

        org_id = token['email']
        access_token = token['accessToken']

        save_auth(org_id=org_id, access_token=access_token)
        print(f"[bold green]Successfully logged in to your account!")

        print('org_id:', org_id)
        print('accessToken:', access_token)

    else:
        print(f"[bold green]Unable to login to your account!")

 

@app.command()
def status():
    print()
    path = PATH_USER_TOKEN

    curr_path = Path(__file__).parent.resolve()
    if os.path.exists(path):
        token = open(path, "r").read()
        public_key = open(f"{curr_path}/public.pem", "rb").read()
        payload = jwt.decode(token, public_key, algorithms=["RS256"])
        print(f"[bold green]You are currently logged in as {payload['email']}")
    else:
        print("[bold red]You are not logged in!")

def statusHelper():
    path = PATH_USER_TOKEN

    if os.path.exists(path):
        return open(path, "r").read()
    else:
        return None

def auth_validate():
    token = statusHelper()
    if not token:
        print("[bold red]You are not logged in!")
        raise typer.Exit()
    return token

@app.command()
def logout():
    print()
    path = PATH_USER_TOKEN

    if os.path.exists(path):
        os.remove(path)
        print(f"[bold yellow]Successfully logged out!")
    else:
        print(f"[bold red]You are not logged in!")

if __name__ == "__main__":
    app()



import requests
import typer
from rich import print
from rich.syntax import Syntax

import os 
import json


from . import get_token, get_project_id, get_org_id

from pureml.utils.constants import BASE_URL, PATH_DATASET_DIR
from urllib.parse import urljoin
import joblib
import pandas as pd
from pureml.utils.hash import check_hash_status_dataset


def save_dataset(dataset, name:str):
    # file_name = '.'.join([name, 'parquet'])
    file_name = '.'.join([name, 'pkl'])
    save_path = os.path.join(PATH_DATASET_DIR, file_name)


    os.makedirs(PATH_DATASET_DIR, exist_ok=True)

    # dataset.to_parquet(save_path)
    joblib.dump(dataset, save_path)

    
    return save_path




def list():
    '''This function will return a list of all the datasets in the project
    
    Returns
    -------
        A list of all the datasets in the project
    
    '''

    user_token = get_token()
    org_id = get_org_id()
    project_id = get_project_id()

    url_path_1 = '{}/project/{}/datasets'.format(org_id, project_id)
    url = urljoin(BASE_URL, url_path_1)
    
    data = {"project_id": project_id}

    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': 'Bearer {}'.format(user_token)
    }


    response = requests.get(url,data=data, headers=headers)

    if response.status_code == 200:

        response_text = response.json()
        dataset_list = response_text['data']

        return dataset_list
    else:
        print(f"[bold red]Unable to obtain the list of datasets!")
        
    return


# def create(name:str):

#     user_token = get_token()
#     org_id = get_org_id()
#     project_id = get_project_id()


#     url_path_1 = '{}/project/{}/dataset/create'.format(org_id, project_id)
#     url = urljoin(BASE_URL, url_path_1)


#     headers = {
#         'Authorization': 'Bearer {}'.format(user_token)
#     }

#     data = {'name': name, 'project_id':project_id}

#     dataset_details = details(name=name)

#     if dataset_details is not None:
#         print("[bold yellow] Dataset already exists")
#         # response_text = response.json()
#         # dataset_details = response_text['data'][0]

#         return dataset_details
#     else:

#         response = requests.post(url, data=data, headers=headers)

#         if response.status_code == 200:
#             print(f"[bold green]Dataset has been Created!")
#             response_text = response.json()
#             dataset_uploaded_details = response_text['data'][0]

#             return dataset_uploaded_details
#         else:
#             print(f"[bold red]Dataset has not been Created!")
#             # print(response.status_code)
#             # print(response.text)

#             return



def check_dataset_hash(hash: str, name:str):

    user_token = get_token()
    org_id = get_org_id()
    project_id = get_project_id()


    url_path_1 = '{}/project/{}/dataset/{}/hash_status'.format(org_id, project_id, name)
    url = urljoin(BASE_URL, url_path_1)


    headers = {
        'Authorization': 'Bearer {}'.format(user_token)
    }


    data = {'hash': hash, 'project_id':project_id}
    response = requests.post(url, data=data, headers=headers)

    hash_exists = False

    if response.status_code == 200:

        hash_exists = response.json()['data']


    return hash_exists



def register(dataset, name:str, pipeline) -> str:
    ''' The function takes in a dataset, a name and a version and saves the dataset locally, then uploads the
    dataset to the PureML server
    
    Parameters
    ----------
    dataset
        The dataset you want to register
    name : str
        The name of the dataset.
    version: str, optional
        The version of the dataset.
    
    '''
        
    user_token = get_token()
    org_id = get_org_id()
    project_id = get_project_id()

    
    dataset_path = save_dataset(dataset, name)
    name_with_ext = dataset_path.split('/')[-1]
    
    if dataset is not None:


        dataset_exists_locally, dataset_hash =  check_hash_status_dataset(file_path = dataset_path, name=name, item_key='dataset')
        dataset_exists_remote = check_dataset_hash(hash=dataset_hash, name=name)
    else:
        dataset_hash = ''
        dataset_exists_remote = False


    if dataset_exists_remote:

        print(f"[bold red]Dataset already exists. Not registering a new version!")
        return True, dataset_hash, 'latest'
    else:

        url_path_1 = '{}/project/{}/dataset/register'.format(org_id, project_id)
        url = urljoin(BASE_URL, url_path_1)

        headers = {
            'Authorization': 'Bearer {}'.format(user_token)
        }


        files = {'file': (name_with_ext, open(dataset_path, 'rb'))}

        pipeline = json.dumps(pipeline)

        data = {'name': name, 'project_id':project_id, 'hash':dataset_hash, 'pipeline': pipeline}
        response = requests.post(url, files=files, data=data, headers=headers)

        if response.status_code == 200:
            if dataset_hash == '':
               print(f"[bold green]Pipeline has been registered!")
            else:            
                print(f"[bold green]Dataset and pipeline have been registered!")

            dataset_version = response.json()['data'][0]['version']

            return True, dataset_hash, dataset_version
        else:
            print(f"[bold red]Dataset has not been registered!")
            
            return True, dataset_hash, None



def details(name:str, version:str='latest'):
    '''It fetches the details of a dataset.
    
    Parameters
    ----------
    name : str
        The name of the dataset
    version: str
        The version of the dataset
    Returns
    -------
        The details of the dataset.
    
    '''
    
    user_token = get_token()
    org_id = get_org_id()
    project_id = get_project_id()

    url_path_1 = '{}/project/{}/dataset/{}/{}/details'.format(org_id, project_id, name, version)
    url = urljoin(BASE_URL, url_path_1)

    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': 'Bearer {}'.format(user_token)
    }
    

    response = requests.get(url, headers=headers)


    if response.status_code == 200:
        # print(f"[bold green]Dataset details have been fetched")
        response_text = response.json()
        dataset_details = response_text['data'][0]
        
        return dataset_details

    else:
        print(f"[bold red]Dataset details have not been found")
        return 



def fetch(name:str, version:str='latest'):
    '''This function fetches a dataset from the server and returns it as a dataframe object
    
    Parameters
    ----------
    name : str, optional
        The name of the dataset you want to fetch.
    version: str
        The version of the dataset
    
    Returns
    -------
        The dataset dataframe is being returned.
    
    '''

    user_token = get_token()
    org_id = get_org_id()
    project_id = get_project_id()


    dataset_details = details(name=name, version=version)

    if dataset_details is None:
        print(f"[bold red]Unable to fetch Dataset")
        return


    # dataset_location = dataset_details['location']
    # dataset_url = 'https://{}'.format(dataset_location)

    dataset_url = dataset_details['location']
    dataset_url = dataset_url.replace('https://pureml-registry.67f23f4479798ac96c01212517a90146.r2.cloudflarestorage.com', 
                                        'https://pub-072ac07f18cd4246bd5c879e7a9df94e.r2.dev')


    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': 'Bearer {}'.format(user_token)
    }
    
    # print('url', dataset_url)

    # response = requests.get(dataset_url, headers=headers)

    response = requests.get(dataset_url)

    if response.status_code == 200:
        dataset_bytes = response.content
        # open('temp_dataset.parquet', 'wb').write(dataset_bytes)
        # dataset = pd.read_parquet('temp_dataset.parquet')

        open('temp_dataset.pkl', 'wb').write(dataset_bytes)
        dataset = joblib.load('temp_dataset.pkl')


        # print(f"[bold green]Dataset has been fetched")
        return dataset
    else:
        print(f"[bold red]Unable to fetch Dataset")
        # print(response.status_code)
        # print(response.text)
        # print(response.url)
        return 




def delete(name:str, version:str='latest') -> str:
    ''' This function deletes a dataset from the project
    
    Parameters
    ----------
    name : str
        The name of the dataset you want to delete
    version : str
        The version of the dataset to delete.
    
    '''

    user_token = get_token()
    org_id = get_org_id()
    project_id = get_project_id()


    url_path_1 = '{}/project/{}/dataset/{}/delete'.format(org_id, project_id, name)
    url = urljoin(BASE_URL, url_path_1)

    
    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': 'Bearer {}'.format(user_token)
    }

    data = {'version':version}
    
    
    response = requests.delete(url, headers=headers, data=data)


    if response.status_code == 200:
        print(f"[bold green]Dataset has been deleted")
        
    else:
        print(f"[bold red]Unable to delete Dataset")

    return response.text







from pureml.utils.pipeline import add_transformer_to_config



def transformer(name:str, parent:str=None):

    def decorator(func):
        # print('Inside decorator')

        # print('decorating', func, 'with argument', name)
        
        def wrapper(*args, **kwargs):
            # print("Inside wrapper")
            
            func_output = func(*args, **kwargs)

            add_transformer_to_config(name=name, func=func, parent=parent)

            res_text = ''


            return func_output

        # print("Outside  wrapper")

        return wrapper
    # print('Outside decorator')
        
    return decorator


import os
import requests
import shutil
from rich import print
import os 
from . import get_token, get_project_id, get_org_id, get_project_name

from pureml.utils.constants import BASE_URL, PATH_USER_PROJECT, PATH_USER_PROJECT_DIR
from urllib.parse import urljoin
import json
from pureml.utils.config import load_config

# os.makedirs(PATH_USER_PROJECT_DIR, exist_ok=True)

def details(name: str=None, id:str = None):
    '''It takes a project_id as input and returns the details of the project if it exists in the remote
    database
    
    Parameters
    ----------
    name : str
        The name of the project to get details for.
    id : str
        The ID of the project to get details for.
    
    '''


    user_token = get_token()
    org_id = get_org_id()

    #If both project name and project id are given, preference will be given to project id
    if name is not None:
        url_path_1 = '{}/project/name/{}'.format(org_id, name)
    else:
        if id is not None:
            url_path_1 = '{}/project/id/{}'.format(org_id, id)
        else:
            return

    
    
    
    url = urljoin(BASE_URL, url_path_1)


    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': 'Bearer {}'.format(user_token)
    }

    response = requests.get(url,headers=headers)

    return response


def save_project(response:requests.Response):

    if os.path.exists(PATH_USER_PROJECT_DIR):
        shutil.rmtree(PATH_USER_PROJECT_DIR)
        os.makedirs(PATH_USER_PROJECT_DIR, exist_ok=True)

        config = load_config()

    with open(PATH_USER_PROJECT, "w") as f:
        project_details = response.json()
        project_details = project_details['data'][0]
        project_details = json.dumps(project_details)
        f.write(project_details)
    



# @app.command()
def init(name:str, description:str=''):
    '''It creates a project if it doesn't exist
    
    Parameters
    ----------
    name : str
        The name of the project.
    description : str
        The description of the project.
    
    '''


    user_token = get_token()
    org_id = get_org_id()
    
    url_path_1 = '{}/project/create'.format(org_id)
    url = urljoin(BASE_URL, url_path_1)


    print("Project Name: ", name)
    print("Description:", description)
    


    data = {"name": name, "description": description}

    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': 'Bearer {}'.format(user_token)
    }

        
    project_response = details(name=name)
    if project_response is None:
        print('Incorect Project Id or Name')
    else:
    
        if project_response.status_code == 200:
            print("[bold yellow] Connected to project.")
            save_project(response = project_response)

        else:
            response = requests.post(url, data=data, headers=headers)

            if response.status_code == 200:
                print("[bold green] Initialized the project.")
                save_project(response = response)
            else:
                # print(response.text)
                print('[bold red] Unable to create Project')

                if os.path.exists(PATH_USER_PROJECT_DIR):
                    shutil.rmtree(PATH_USER_PROJECT_DIR)
                    os.makedirs(PATH_USER_PROJECT_DIR, exist_ok=True)

                    config = load_config()


def delete(id:str):
    '''It deletes a project
    
    Parameters
    ----------
    project_id : str
        The id of the project you want to delete.
    
    '''


    user_token = get_token()
    org_id = get_org_id()

    

    url_path_1 = '{}/project/{}/delete'.format(org_id, id)
    url = urljoin(BASE_URL, url_path_1)


    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': 'Bearer {}'.format(user_token)
    }

    def delete_project(url, headers):

        response = requests.delete(url, headers=headers)
        if response.status_code == 200:
            print(f"[bold green] Deleted the project!")
            # print(response.text)
        else:
            print(f"[bold red] Could not delete the project!")
            # print(response.status_code)
            # print(response.text)

        return response


    reponse_details = details(id=id)

    if reponse_details is None:
        print('Incorect Project Id or Name')
    else:
        if reponse_details.status_code == 200:        
            reponse_delete = delete_project(url=url, headers=headers)
        
        else:
            print(f"[bold red] Project doesnot exists!")
            # return




    

def list():
    '''`list()` is a function that calls the API to obtain a list of all projects
    
    Returns
    -------
        The response object is being returned.
    
    '''


    user_token = get_token()
    org_id = get_org_id()

    url_path_1 = '{}/project/all'.format(org_id)
    url = urljoin(BASE_URL, url_path_1)

    
    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': 'Bearer {}'.format(user_token)
    }
    
    response = requests.get(url, headers=headers)

    print(response.elapsed.total_seconds())
    
    if response.status_code == 200:
        # print(f"[bold green] Obtained the project list")
        response_text = response.json()
        project_list = response_text['data']
        # print(project_list)

        return project_list
    else:
        print(f"[bold red] Unable to obtain the project list!")
        return
    


import os
from pathlib import Path
from .packaging import Model



def save_model(model, model_name, model_path) -> None: 
    ''' The function takes in a model and a model name, and saves the model to the default `models` directory. 
    
    Parameters
    ----------
    model
        The model you want to save.
    model_name
        The name of the model.
    
    '''
    model_dir = os.path.dirname(model_path)
    print(model_dir)
    # os.makedirs(model_dir, exist_ok=True)

    Model(model=model, model_name=model_name, model_path=model_path).save_model()



def load_model(model_path:Path):
    ''' Loads a model from a given path
    
    Parameters
    ----------
    model_path : Path
        The path to the model you want to load.
    
    Returns
    -------
        The model is being returned.
    
    '''
    model = Model(model_path=model_path).load_model()

    return model




from pureml.components.model import register
from pureml.utils.pipeline import add_model_to_config, load_metrics_from_config, load_params_from_config
from pureml import metrics, params


def model(name:str):
    
    def decorator(func):
        # print('Inside decorator')

        # print('Adding model name: ', name, 'to config before invoking user function')
        add_model_to_config(name=name)
        
        def wrapper(*args, **kwargs):
            # print("Inside wrapper")
            
            func_output = func(*args, **kwargs)

            model_exists_in_remote, model_hash, model_version = register(model=func_output, name=name)

            if model_exists_in_remote:                 #Only add the model to config if it is successfully pushed

                add_model_to_config(name=name, hash=model_hash, version=model_version, func=func)

                metric_values = load_metrics_from_config()
                if len(metric_values) !=0 :
                    metrics.add(metrics=metric_values, model_name=name, model_version=model_version)

                param_values = load_params_from_config()
                if len(param_values) !=0 :
                    params.add(params=param_values, model_name=name, model_version=model_version)
 

            return func_output


        # print("Outside  wrapper")

        return wrapper
    # print('Outside decorator')
        
    return decorator


import pandas as pd
import numpy as np
from PIL import Image
import json


def process_input(input):

    input_type = ''
    input_shape = None


    if input is not None:
        input_keys = input.keys()

        if 'type' in input_keys:
            input_type = input['type']
        if 'shape' in input_keys:
            input_shape = input['shape']
    
    return input_type, input_shape



def process_output(output):

    output_type = ''
    output_shape = None


    if output is not None:
        output_keys = output.keys()

        if 'type' in output_keys:
            output_type = output['type']
        if 'shape' in output_keys:
            output_shape = output['shape']
    
    return output_type, output_shape



def parse_input(data, input_type, input_shape):
    # if input_type == 'json':
    #     data = data

    if input_type == 'numpy ndarray':
        if type(data) == str:
            data = json.loads(data)
        data = np.array(data)
        data = data.reshape(input_shape)
    elif input_type == 'pandas dataframe':
        if type(data) == str:
            data = json.loads(data)
        data =  pd.DataFrame.from_dict(data)
    elif input_type == 'text':
        data = json.loads(data)
    elif input_type == 'image':
        data = Image.open(data)
        data = np.array(data)
        print(data.shape)
    else:
        data = None

    print(type(data))
    # print(data)

    return data

def parse_output(data, output_type, output_shape):
    # if input_type == 'json':
    #     data = data


    if output_type == 'numpy ndarray':
        data = data.tolist()
        data = json.dumps(data)
    elif output_type == 'pandas dataframe':
        data = data.to_dict()
        data = json.dumps(data)
    elif output_type == 'text':
        data = json.dumps(data)
    else:
        data = json.dumps(data)



    return data




from pureml.components.dataset import register
from pureml.utils.pipeline import add_dataset_to_config
from pureml.pipeline.data.create_pipeline import create_pipeline

def dataset(name:str, parent:str=None, upload=False):

    def decorator(func):

        #Add dataset name to config here if it is being used by any of the pipeline components.
        # add_dataset_to_config(name=name, parent=parent)
        
        def wrapper(*args, **kwargs):
            
            func_output = func(*args, **kwargs)

            if not upload:
                dataset = None
            else:
                dataset = func_output


            add_dataset_to_config(name=name, parent=parent, func=func)

            pipeline = create_pipeline()


            dataset_exists_in_remote, dataset_hash, dataset_version = register(dataset=dataset, name=name, pipeline=pipeline)

            #Uncomment this if there any components that depend on dataset version, or dataset hash
            # if dataset_exists_in_remote:
            #     add_dataset_to_config(name=name, hash=dataset_hash, version=dataset_version, parent=parent, func=func)

            return func_output

        return wrapper

        
    return decorator


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pureml',
 'pureml.cli',
 'pureml.components',
 'pureml.config',
 'pureml.decorators',
 'pureml.deploy',
 'pureml.packaging',
 'pureml.packaging.model_packaging',
 'pureml.pipeline',
 'pureml.pipeline.data',
 'pureml.utils']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=9.3.0,<10.0.0',
 'PyJWT>=2.4.0,<3.0.0',
 'PyYAML>=6.0,<7.0',
 'docker>=6.0.1,<7.0.0',
 'fastapi>=0.88.0,<0.89.0',
 'joblib>=1.2.0,<2.0.0',
 'matplotlib>=3.6.2,<4.0.0',
 'numpy>=1.23.1,<2.0.0',
 'pandas>=1.4.3,<2.0.0',
 'pyarrow>=8.0.0,<9.0.0',
 'pydantic>=1.9.1,<2.0.0',
 'python-dotenv>=0.20.0,<0.21.0',
 'python-multipart>=0.0.5,<0.0.6',
 'requests>=2.28.1,<3.0.0',
 'typer[all]>=0.6.1,<0.7.0',
 'uvicorn>=0.20.0,<0.21.0']

entry_points = \
{'console_scripts': ['pureml = pureml.cli.main:app']}

setup_kwargs = {
    'name': 'pureml',
    'version': '0.1.6',
    'description': '',
    'long_description': '<h1 align="center">\n  <a href="https://pureml.com">\n    <img\n      align="center"\n      alt="PureML"\n      src="https://github.com/PureML-Inc/PureML/blob/readme/assets/coverImg.jpeg"\n      style="width:100%;"\n    />\n  </a>\n</h1>\n\n\n\n\n<div align="center">\n\n# Track, version, compare and review your data and models.\n\n</div>\n\n\n# Quick Access\n\n<p align="center">\n  <a\n    href="https://docs.pureml.com"\n  ><b>Documentation</b></a>\n  &nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;\n  <a\n    href="https://www.youtube.com/watch?v=HdzLFEWS4s8&t=1s"\n  ><b>Watch Demo</b></a>\n  &nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;\n  <a\n    href="https://docs.pureml.com/docs/get-started/quickstart_tabular"\n  ><b>Quick example</b></a>\n  &nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;\n  <a\n    href="#"\n  ><b>Get Instant Help</b></a>\n  &nbsp;&nbsp;&nbsp; &nbsp;&nbsp;&nbsp;\n  <a\n    href="https://app.pureml.com/auth/signup"\n  ><b>Sign Up for free</b></a>\n\n</p>\n\n<!-- \n<img\n  referrerpolicy="no-referrer-when-downgrade"\n  src="https://static.scarf.sh/a.png?x-pxid=b3c96d79-b8f0-414b-a687-8bfc164b4b7a"\n/> -->\n\n</br>\n</br>\n\n\n<div align="center">\n  <a\n    href="https://pypi.org/project/pureml/"\n  >\n    <img alt="PyPi" src="https://img.shields.io/pypi/v/pureml?color=green&logo=pureml" />\n  </a>\n  <a\n    href="https://python-poetry.org/"\n  >\n    <img src="https://img.shields.io/badge/poetry-1.1.14-blue?style=flat&logo=poetry&logoColor=white" />\n  </a>\n  <a\n    href="https://opensource.org/licenses/Apache-2.0"\n  >\n    <img alt="License" src="https://img.shields.io/pypi/l/pureml?color=red&logo=Apache&logoColor=red" />\n  </a>\n  <a\n    href="https://discord.gg/xNUHt9yguJ"\n  >\n    <img alt="Discord" src="https://img.shields.io/badge/Discord-Join%20Discord-blueviolet?style=flat&logo=discord&logoColor=white" />\n  </a>\n  <a\n    href="https://pepy.tech/project/pureml"\n  >\n    <img alt="Downloads" src="https://static.pepy.tech/badge/pureml">\n  </a>\n  <a\n    href="https://pypi.org/project/pureml/"\n  >\n    <img alt="^3.8" src="https://img.shields.io/pypi/pyversions/pureml">\n  </a>\n  <a\n    href="https://pypi.org/project/pureml/"\n  >\n    <img alt="Coverage" src="https://img.shields.io/codecov/c/github/PureML-Inc/pureml">\n  </a>\n\n</div>\n\n\n</br>\n</br>\n\n\n\n# Intro\n\nPureML is an open-source version control for machine learning.\n\n1. [Quick start](#quick-start)\n1. [How it works](#how-it-works)\n1. [Demo](#demo)\n1. [Main Features](#main-features)\n1. [Tutorials](#tutorials)\n1. [Core design principles](#core-design-principles)\n1. [Core abstractions](#core-abstractions)\n1. [Why to get involved](#why-to-get-involved)\n\n<br />\n\n# Quick start\n\nYou can install and run PureML using `pip`.\n\n\n### Using `pip`\n\n1. Install PureML\n    ```bash\n    pip install pureml\n    ```\n\n<br />\n\n# How it works\nJust add a few lines of code. You don\'t need to change the way you work.\n\nPureML is a Python library that uploads metadata to S3.\n\n### Generating Data Lineage\n\n1. Load Data\n```python\n@load_data(name=\'loading data\')\ndef loading_data():\n    \n    return pd.read_csv(\'churn.csv\')\n```\n\n2. Transform Data\n```python\n@transformer(name=\'fill missing values\')\ndef fill_missing_values(df):\n    return df.fillna()\n    \n\n@transformer(name=\'encode ordinal\')\ndef encode_ordinal(df):\n    col_ord = [\'state\', \'phone number\']\n    df_ord = df[col_ord]\n    feat = OrdinalEncoder().fit_transform(df_ord)    \n    df[col_ord] = feat\n    \n    return df\n\n@transformer(name=\'encode binary\')\ndef encode_binary(df):\n\n    df[\'voice mail plan\'] = df[\'voice mail plan\'].map({\'yes\':1, \'no\':0})\n    df[\'international plan\'] = df[\'international plan\'].map({\'yes\':1, \'no\':0})\n    df[\'churn\'] = df[\'churn\'].map({True:1, False:0})\n\n    return df\n```\n\n3. Register Dataset\n```python\n@dataset(name=\'telecom churn\', parent=\'encode binary\')\ndef build_dataset():\n    df = loading_data()\n\n    df = fill_missing_values(df)\n\n    df = encode_ordinal(df)\n\n    df = encode_binary(df)\n\n    return df\n\ndf = build_dataset()\n```\n\nThis is how generated data lineage will look like in the UI\n\n<h1 align="center">\n    <img\n      align="center"\n      src="https://github.com/PureML-Inc/PureML/blob/readme/assets/pipeline.png?raw=true"\n      style="width:60%;"\n    />\n  </a>\n</h1>\n\n\n# Demo\n\n### Live demo\n\nBuild and run a PureML project to create data lineage and a model with our <b>[demo colab link](https://colab.research.google.com/drive/1LlrpaKiREwgesaRcnwkJP-w2MPesXf1t?usp=sharing)</b>.\n\n\n### Demo video (2 min)\nPureML quick start demo\n\n[![PureML Demo Video](https://github.com/PureML-Inc/PureML/blob/readme/assets/demo_video_cover.png?raw=true)](https://www.youtube.com/watch?v=HdzLFEWS4s8&ab_channel=PureMLInc.)\n\n<!-- <iframe\n    width="640"\n    height="480"\n    src="https://www.youtube.com/watch?v=HdzLFEWS4s8&ab_channel=PureMLInc."\n    frameborder="0"\n    allow="autoplay; encrypted-media"\n    allowfullscreen\n> -->\n</iframe>\n\n<sub><i>Click the image to play video</i></sub>\n\n<br />\n\n\n# [Main Features](https://docs.pureml.com/)\n|   |   |\n| --- | --- |\n| Data Lineage | Automatic generation of data lineage|\n| Dataset Versioning | Automatic Semantic Versioning of datasets |\n| Model Versioning | Automatic Semantic Versioning of models |\n| Comparision | Comparing different versions of models or datasets\n| Branches (*Coming Soon*) | Separation between experimentation and production ready models using branches |\n| Review (*Coming Soon*) | Review and approve models, and datasets to production ready branch|\n\n<br />\n\n\n# Tutorials\n\n- [Registering Data lineage](https://docs.pureml.com/docs/data/register_data_pipeline)\n- [Registering models](https://docs.pureml.com/docs/models/register_models)\n- [Quick Start: Tabular](https://docs.pureml.com/docs/get-started/quickstart_tabular)\n- [Quick Start: Computer Vision](https://docs.pureml.com/docs/get-started/quickstart_cv)\n- [Quick Start: NLP](https://docs.pureml.com/docs/get-started/quickstart_nlp)\n- [Logging](https://docs.pureml.com/docs/log/overview)\n\n\n<br />\n\n\n# Core design principles\n\n|   |   |\n| --- | --- |\n| Easy developer experience | An intuitive open source package aimed to bridge the gaps in data science teams |\n| Engineering best practices built-in | Integrating PureML functionalities in your code doenot disrupt your workflow |\n| Object Versioning | A reliable object versioning mechanism to track changes to your datasets, and models |\n| Data is a first-class citizen | Your data is secure. It will never leave your system. |\n| Reduce Friction | Have access to operations performed on data using data lineage without having to spend time on lengthy meetings |\n\n\n\n<br />\n\n# Core abstractions\n\nThese are the fundamental concepts that PureML uses to operate.\n\n|   |   |\n| --- | --- |\n| [Project](https://docs.pureml.com/docs/projects/about_projects) | A data science project. This is where you store datasets, models, and their related objects. It is similar to a github repository with object storage.|\n| [Lineage](https://docs.pureml.com/docs/data/register_data_pipeline) | Contains a series of transformations performed on data to generate a dataset.|\n| Data Versioning | Versioning of the data should be comprehensible to the user and should encapsulate the changes in the data, its creation mechanism, among others.|\n| Model Versioning| Versioning of the model should be comprehensible to the user and should encapuslate the changes in training data, model architecture, hyper parameters.|\n| Fetch | This functionality is used to fetch registered Models, and Datasets.|\n\n\n<br />\n\n# Why to get involved\nVersion control is much more common in software than in machine learning. So why isn’t everyone using Git? Git doesn’t work well with machine learning. It can’t handle large files, it can’t handle key/value metadata like metrics, and it can’t record information automatically from inside a training script.\n\nGitHub wasn’t designed with data as a core project component. This along with a number of other differences between AI and more traditional software projects makes GitHub a bad fit for artificial intelligence, contributing to the reproducibility crisis in machine learning.\n\nFrom manually tracking models to git based versioning systems that do not follow an intuitive versioning mechanism, there is no standardized way to track objects. Using these mechanisms, it is hard enough to track or get your model from a month ago running, let alone of a teammates!\n\nWe are trying to build a version control system for machine learning objects. A mechanism that is object dependant and intuitive for users.\n\nLets build this together. If you have faced this issue or have worked out a similar solution for yourself, please join us to help build a better system for everyone.\n\n<br />\n\n# Reporting Bugs\nTo report any bugs you have faced while using PureML package, please\n1. report it in [Discord](https://discord.gg/xNUHt9yguJ) channel\n1. Open an [issue](https://github.com/PureML-Inc/PureML/issues)\n\n<br />\n\n# Contributing and developing\nLets work together to improve the features for everyone. \n\nWork with mutual respect.\n\n\n<br />\n\n# License\nSee the [Apache-2.0](./License) file for licensing information.\n\n\n\n<br />\n',
    'author': 'vamsidhar muthireddy',
    'author_email': 'vamsi.muthireddy@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pureml.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
